# -*- coding: utf-8 -*-
# Copyright (c) 2020 Richard Hull and contributors
# See LICENSE.rst for details.

from pathlib import Path
from math import ceil
from copy import deepcopy
from PIL import Image, ImageFont
import cbor2
from luma.core.util import from_16_to_8, from_8_to_16


class bitmap_font():
    """
    An PIL.Imagefont style font

    The structure of this class was modeled after the PIL ImageFont class
    and is intended to be interchangable for PIL.ImageFont objects.  It has
    the following additional capabilities.

    * Allows fonts larger than 256 characters to be created
    * Font can be combined with other fonts
    * Font characters can be mapped to their correct unicode codepoints
    * Font can be initialized from a basic sprite table (with conditions)

    .. note:
        Because this font is implemented completely in Python it will be slower
        than a native PIL.ImageFont object.

    .. versionadded:: 1.16.0
    """

    PUA_SPACE = 0xF8000

    def __init__(self):
        self.mappings = {}
        self.metrics = []

    def load(self, filename):
        """
        Load font from filename

        :param filename: the filename of the file containing the font data
        :type filename: str
        :return: a font object
        :rtype: luma.core.bitmap_font
        """
        with open(filename, 'rb') as fp:
            s = fp.readline()
            if s != b'LUMA.CORE.BITMAP_FONT\n':
                raise SyntaxError('Not a luma.core.bitmap_font file')
            fontdata = cbor2.load(fp)

        self._load_fontdata(fontdata)
        return self

    def loads(self, fontdata):
        """
        Load luma.core.bitmap_font from a string of serialized data produced
        by the dumps method

        :param fontdata: The serialized font data that will be used to initialize
            the font.  This data is produced by the luma.core.bitmap_font.dumps
            method.
        :type fontdata: bytes
        :return: a font object
        :rtype: luma.core.bitmap_font
        """
        fontdata = cbor2.loads(fontdata)
        self._load_fontdata(fontdata)
        return self

    def load_pillow_font(self, file, mappings=None):
        """
        Create luma.core.bitmap_font from a PIL ImageFont style font.

        :param file: The filename of the PIL.ImageFont to load
        :type file: str
        :param mappings: a dictionary of unicode to value pairs (optional).
            Mappings allow the appropriate unicode values to be provided for
            each character contained within the font
        :type mappings: dict
        :return: a font object
        :rtype: luma.core.bitmap_font
        """
        with open(file, 'rb') as fp:
            if fp.readline() != b"PILfont\n":
                raise SyntaxError("Not a PIL.ImageFont file")
            while True:
                s = fp.readline()
                if not s:
                    raise SyntaxError("PIL.ImageFont file missing metric data")
                if s == b"DATA\n":
                    break
            data = fp.read(256 * 20)
            if len(data) != 256 * 20:
                raise SyntaxError("PIL.ImageFont file metric data incomplete")

        sprite_table = self._get_image(file)
        self._populate_metrics(sprite_table, data, range(256), mappings)
        sprite_table.close()
        return self

    def load_sprite_table(self, sprite_table, index, xwidth, glyph_size, cell_size, mappings=None):
        """
        Load a font from a sprite table

        :param sprite_table: A PIL.Image representation of every glyph within the font
        :type sprite_table: PIL.Image
        :param index: The list of character values contained within sprite_table.
            This list MUST be in the same order that the glyphs for the characters
            appear within the sprite_table (in left to right, top to bottom order)
        :type index: list or other iterable
        :param xwidth: number of pixels between placements of each character in a
            line of text
        :type xwidth: int
        :param glyph_size: tuple containing the width and height of each character
            in the font
        :type glyph_size: tuple(int, int)
        :param cell_size: tuple containing the width and height of each cell in the
            sprite table.  Defaults to the size of the glyphs.
        :type cell_size: tuple(int, int)
        :param mappings: a dictionary of unicode to value pairs (optional)
            Mappings allow the appropriate unicode values to be provided for
            each character contained within the font
        :type mappings: dict
        :return: a font object
        :rtype: luma.core.bitmap_font

        .. note:
            Font contained within table must adhere to the following conditions
            * All glyphs must be the same size
            * Glyphs are contained within the sprite table in a grid arrangement
            * The grid is filled with glyphs placed in horizontal order
            * Each cell in the grid is the same size
            * The placement of each glyph has no offset from its origin
        """

        table_width = sprite_table.size[0]

        # Each character uses the same data
        line = [xwidth, 0, 0, -glyph_size[1], glyph_size[0], 0]

        data = []
        # Generate an entry for each character
        for c in range(len(index)):
            offset = c * cell_size[0]
            left = offset % table_width
            top = (offset // table_width) * cell_size[1]
            right = left + glyph_size[0]
            bottom = top + glyph_size[1]
            data = data + line + [left, top, right, bottom]
        self._populate_metrics(sprite_table, from_16_to_8(data), index, mappings)
        return self

    def save(self, filename):
        """
        Write luma.core.bitmap_font data to a file
        """
        with open(filename, 'wb') as fp:
            fontdata = self._generate_fontdata()
            fp.write(b'LUMA.CORE.BITMAP_FONT\n')
            cbor2.dump(fontdata, fp)

    def dumps(self):
        """
        Serializes the font data for transfer or storage

        :return: serialized font data
        :rtype: bytes
        """
        fontdata = self._generate_fontdata()
        return cbor2.dumps(fontdata)

    def _generate_fontdata(self):
        """
        Utility method to create an efficient serializable representation
        of a luma.core.bitmap_font
        """
        cell_size = (self.width, self.height)

        area = self.count * self.width * self.height
        table_width = ((ceil(area**0.5) + self.width - 1) // self.width) * self.width
        table_height = ((ceil(area / table_width) + self.height - 1) // self.height) * self.height

        image = Image.new('1', (table_width, table_height))
        metrics = []
        for i, v in enumerate(self.metrics):
            offset = i * cell_size[0]
            left = offset % table_width
            top = (offset // table_width) * cell_size[1]
            image.paste(v['img'], (left, top))
            metrics.append((v['xwidth'], v['dst']))

        fontdata = {}
        fontdata['type'] = 'LUMA.CORE.BITMAP_FONT'
        fontdata['count'] = len(self.metrics)
        if self.regular:
            fontdata['xwidth'] = self.regular[0]
            fontdata['glyph_size'] = (self.width, self.height)
        fontdata['cell_size'] = cell_size
        fontdata['mappings'] = self.mappings
        fontdata['sprite_table_dimensions'] = (table_width, table_height)
        fontdata['sprite_table'] = image.tobytes()
        if not self.regular:
            fontdata['metrics'] = metrics

        return fontdata

    def _load_fontdata(self, fontdata):
        """
        Initialize font from deserialized data
        """
        try:
            count = fontdata['count']
            xwidth = fontdata.get('xwidth')
            glyph_size = fontdata.get('glyph_size')
            cell_size = fontdata['cell_size']
            self.mappings = fontdata['mappings']
            table_width, table_height = fontdata['sprite_table_dimensions']
            image = Image.frombytes('1', (table_width, table_height), fontdata['sprite_table'])
            metrics = fontdata.get('metrics')
        except (KeyError, TypeError, ValueError):
            raise ValueError('Cannot parse fontdata. It is invalid.')

        self.metrics = []
        for i in range(count):
            offset = i * cell_size[0]
            metric = metrics[i] if metrics else (xwidth, [0, -glyph_size[1], glyph_size[0], 0])
            left = offset % table_width
            top = (offset // table_width) * cell_size[1]
            right = left + (metric[1][2] - metric[1][0])
            bottom = top + (metric[1][3] - metric[1][1])

            self.metrics.append({
                'xwidth': metric[0],
                'dst': metric[1],
                'img': image.crop((left, top, right, bottom))
            })
        self._calculate_font_size()

    def _get_image(self, filename):
        """
        Load sprite_table associated with font
        """
        ifs = {p.resolve() for p in Path(filename).parent.glob(Path(filename).stem + ".*") if p.suffix in (".png", ".gif", ".pbm")}
        for f in ifs:
            try:
                image = Image.open(f)
            except:
                pass
            else:
                if image.mode in ['1', 'L']:
                    break
                image.close()
        else:
            raise OSError('cannot find glyph data file')
        return image

    def _lookup(self, val):
        """
        Utility method to determine a characters placement within the metrics list
        """
        if val in self.mappings:
            return self.mappings[val]
        if val + self.PUA_SPACE in self.mappings:
            return self.mappings[val + self.PUA_SPACE]
        return None

    def _getsize(self, text):
        """
        Utility method to compute the rendered size of a line of text.  It
        also computes the minimum column value for the line of text. This is
        needed in case the font has a negative horizontal offset which
        requires that the size be expanded to accomodate the extra pixels.
        """
        min_col = max_col = cp = 0
        for c in text:
            m = self._lookup(ord(c))
            if m is None:
                # Ignore characters that do not exist in font
                continue
            char = self.metrics[m]
            min_col = min(min_col, char['dst'][0] + cp)
            max_col = max(max_col, char['dst'][2] + cp)
            cp += char['xwidth']
        return (max_col - min_col, self.height, min_col)

    def getsize(self, text, *args, **kwargs):
        """
        Wrapper for _getsize to match the interface of PIL.ImageFont
        """
        width, height, min = self._getsize(text)
        return (width, height)

    def getmask(self, text, mode="1", *args, **kwargs):
        """
        Implements an PIL.ImageFont compatible method to return the rendered
        image of a line of text
        """

        # TODO: Test for potential character overwrite if horizontal offset is < 0
        assert mode in ['1', 'L']
        width, height, min = self._getsize(text)
        image = Image.new(mode, (width, height))

        # Adjust start if any glyph is placed before origin
        cp = -min if min < 0 else 0

        for c in text:
            m = self._lookup(ord(c))
            if m is None:
                # Ignore characters that do not exist in font
                continue
            char = self.metrics[m]
            px = char['dst'][0] + cp
            py = char['dst'][1] + self.baseline
            image.paste(char['img'], (px, py))
            cp += char['xwidth']
        return image.im

    def _populate_metrics(self, sprite_table, data, index, mappings):
        """
        Populate metrics on initial font load from a sprite table or PIL ImageFont
        Place characters contained on the sprite_table into Unicode
        private use area (PUA).  Create a reverse lookup from the values
        that are contained on the sprite_table.

        .. note:
            Arbritarily using Supplemental Private Use Area-A starting at
            PUA_SPACE (0xF8000) to give the raw sprite_table locations
            a unicode codepoint.
        """

        self.metrics = []
        self.glyph_index = {}
        self.data = data
        idx = 0

        rev_map = {}
        if mappings is not None:
            for k, v in mappings.items():
                if v in rev_map:
                    rev_map[v].append(k)
                else:
                    rev_map[v] = [k]

        self.mappings = {}
        for i, c in enumerate(index):
            metric = from_8_to_16(data[i * 20:(i + 1) * 20])

            # If character position has no data, skip it
            if sum(metric) == 0:
                continue

            xwidth = metric[0]
            dst = metric[2:6]
            src = metric[6:10]

            img = sprite_table.crop((src[0], src[1], src[2], src[3]))
            self.metrics.append({
                'xwidth': xwidth,
                'dst': dst,
                'img': img
            })
            self.mappings[c + self.PUA_SPACE] = idx
            if c in rev_map:
                for u in rev_map[c]:
                    self.mappings[u] = idx
            i2b = img.tobytes()
            # Only add new glyphs except always add space character
            if i2b not in self.glyph_index or c == 0x20:
                self.glyph_index[i2b] = c
            idx += 1

        self._calculate_font_size()

    def _calculate_font_size(self):
        # Calculate height and baseline of font
        ascent = descent = width = 0

        m = self.metrics[0]
        regular = (m['xwidth'], m['dst'])
        xwidth = regular[0]
        regular_flag = True
        for m in self.metrics:
            if regular != (m['xwidth'], m['dst']):
                regular_flag = False
            ascent = max(ascent, -m['dst'][1])
            descent = max(descent, m['dst'][3])
            width = max(width, m['dst'][2] - m['dst'][0])
            xwidth = max(xwidth, m['xwidth'])
        self.height = ascent + descent
        self.width = width
        self.baseline = ascent
        self.regular = regular if regular_flag else None
        self.count = len(self.metrics)

    def combine(self, source_font, characters=None, force=False):
        """
        Combine two luma.core.bitmap_fonts

        :param source_font: a luma.core.bitmap_font to copy from
        :type: luma.core.bitmap_font
        :param characters: (optional) A list of the characters to transfer from
            the source_font.  If not provided, all of the characters within
            the source_font will be transferred.
        :type characters: str
        :param force: If set, the source_font can overwrite values that already
            exists within this font.  Default is False.
        :type force: bool
        """

        if characters:
            for c in characters:
                if ord(c) in self.mappings and not force:
                    continue
                m = source_font._lookup(ord(c))
                if m is not None:
                    v = source_font.metrics[m]
                else:
                    raise ValueError('{0} is not a valid character within the source font'.format(c))
                self.metrics.append(v)
                self.mappings[ord(c)] = len(self.metrics) - 1
        else:
            # Copy source values into destination but don't overwrite existing characters unless force set
            for k, v in source_font.mappings.items():
                if k in self.mappings and not force:
                    continue
                self.metrics.append(source_font.metrics[v])
                self.mappings[k] = len(self.metrics) - 1

        # Recompute font size metrics
        self._calculate_font_size()


def load(filename):
    """
    Load a luma.core.bitmap_font file.  This function creates a
    luma.core.bitmap_font object from the given luma.core.bitmap_font file, and
    returns the corresponding font object.

    :param filename: Filename of font file.
    :type filename: str
    :return: A luma.core.bitmap_font object.
    :exception OSError: If the file could not be read.
    :exception SyntaxError: If the file does not contain the expected data
    """
    f = bitmap_font()
    f.load(filename)
    return f


def loads(data):
    """
    Load a luma.core.bitmap_font from a string of serialized data.  This function
    creates a luma.core.bitmap object from serialized data produced from the
    dumps method and returns the corresponding font object.

    :param data: Serialized luma.core.bitmap_font data.
    :type data: str
    :return: A luma.core.bitmap_font object.
    :exception ValueError: If the data does not a valid luma.core.bitmap_font
    """
    f = bitmap_font()
    f.loads(data)
    return f


def load_pillow_font(filename, mappings=None):
    """
    Load a PIL font file.  This function creates a luma.core.bitmap_font object
    from the given PIL bitmap font file, and returns the corresponding font object.

    :param filename: Filename of font file.
    :type filename: str
    :param mappings: a dictionary of unicode to value pairs (optional)
    :type mappings: dict
    :return: A font object.
    :exception OSError: If the file could not be read.
    :exception SyntaxError: If the file does not contain the expected data
    """
    f = bitmap_font()
    f.load_pillow_font(filename, mappings)
    return f


def load_sprite_table(sprite_table, index, xwidth, glyph_size, cell_size=None, mappings=None):
    """
    Create a luma.core.bitmap_font from a sprite table.

    :param sprite_table: Filename of a sprite_table file or a PIL.Image containing the
        sprite_table
    :type sprite_table: str or PIL.Image
    :param index: The list of character values contained within sprite_table.
        This list MUST be in the same order that the glyphs for the characters
        appear within the sprite_table (in left to right, top to bottom order)
    :type index: list or other iterable
    :param xwidth: number of pixels between placements of each character in a
        line of text
    :type xwidth: int
    :param glyph_size: tuple containing the width and height of each character
        in the font
    :type glyph_size: tuple(int, int)
    :param cell_size: tuple containing the width and height of each cell in the
        sprite table.  Defaults to the size of the glyphs.
    :type cell_size: tuple(int, int)
    :param mappings: a dictionary of unicode to value pairs (optional)
    :type mappings: dict
    :return: A font object.
    :exception OSError: If the file could not be read.
    :exception SyntaxError: If the file does not contain the expected data

    .. note:
        Requires a font where each character is the same size with no horizontal
        or vertical offset and has consistant horizontal distance between each
        character
    """
    f = bitmap_font()
    need_to_close = False
    if type(sprite_table) == str:
        try:
            sprite_table = Image.open(sprite_table)
            need_to_close = True
        # Differentiate between file not found and invalid sprite table
        except FileNotFoundError:
            raise
        except IOError:
            raise ValueError('File {0} not a valid sprite table'.format(sprite_table))

    if isinstance(sprite_table, Image.Image):
        cell_size = cell_size if cell_size is not None else glyph_size
        f.load_sprite_table(sprite_table, index, xwidth, glyph_size, cell_size, mappings)
    else:
        raise ValueError('Provided image is not an instance of PIL.Image')

    if need_to_close:
        sprite_table.close()
    return f


class embedded_fonts(ImageFont.ImageFont):
    """
    Utility class to manage the set of fonts that are embedded within a
    compatible device.

    :param data: The font data from the device.  See note below.
    :type data: dict
    :param selected_font: The font that should be loaded as this device's
        default.  Will accept the font's index or its name.
    :type selected_font: str or int

    ..note:
        The class is used by devices which have embedded fonts and is not intended
        to be used directly. To initialize it requires providing a dictionary
        of font data including a `PIL.Image.tobytes` representation of a
        sprite_table which contains the glyphs of the font organized in
        consistent rows and columns, a metrics dictionary which provides the
        information on how to retrieve fonts from the sprite_table, and a
        mappings dictionary that provides unicode to table mappings.

    .. versionadded:: 1.16.0
    """
    def __init__(self, data, selected_font=0):
        self.data = data
        self.font_by_number = {}
        self.names_index = {}

        for i in range(len(data['metrics'])):
            name = data['metrics'][i]['name']
            self.names_index[name] = i

        self.current = selected_font

    def load(self, val):
        """
        Load a font by its index value or name and return it

        :param val: The index or the name of the font to return
        :type val: int or str
        """
        if type(val) is str:
            if val in self.names_index:
                index = self.names_index[val]
            else:
                raise ValueError('No font with name {0}'.format(val))
        elif type(val) is int:
            if val in range(len(self.names_index)):
                index = val
            else:
                raise ValueError('No font with index {0}'.format(val))
        else:
            raise TypeError('Expected int or str.  Received {0}'.format(type(val)))

        if index not in self.font_by_number:
            i = index
            index_list = self.data['metrics'][i]['index']
            xwidth = self.data['metrics'][i]['xwidth']
            cell_size = self.data['metrics'][i]['cell_size']
            glyph_size = self.data['metrics'][i]['glyph_size']
            table_size = self.data['metrics'][i]['table_size']
            mappings = self.data['mappings'][i] if 'mappings' in self.data else None
            sprite_table = Image.frombytes('1', table_size, self.data['fonts'][i])
            font = load_sprite_table(sprite_table, index_list, xwidth, glyph_size, cell_size, mappings)
            self.font_by_number[i] = font

        return self.font_by_number[index]

    @property
    def current(self):
        """
        Returns the currently selected font
        """
        return self.font

    @current.setter
    def current(self, val):
        """
        Sets the current font, loading the font if it has not previously been selected

        :param val:  The name or index number of the selected font.
        :type val: str or int
        """
        self.font = self.load(val)

    def combine(self, font, characters=None, force=False):
        """
        Combine the current font with a new one

        :param font: The font to combine with the current font
        :type font: luma.core.bitmap
        :param characters: (Optional) A list of characters to move from the new font to the
            current font.  If not provided all characters from the new font will
            be transferred.
        :type characters: list of unicode characters
        :param force: Determines if conflicting characters should be ignored (default)
            or overwritten.

        .. note:
          This does not permanently change the embedded font.  If you set the value
          of current again even if setting it to the same font, the changes that combine
          has made will be lost.
        """
        destination = deepcopy(self.font)
        destination.combine(font, characters, force)
        self.font = destination



# -*- coding: utf-8 -*-
# Copyright (c) 2017-20 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Encapsulates sending commands and data over a serial interface, whether that
is I²C, SPI or bit-banging GPIO.
"""

import errno
from time import sleep

import luma.core.error
from luma.core import lib


__all__ = ["i2c", "spi", "bitbang", "ftdi_spi", "ftdi_i2c", "pcf8574"]

#: Default amount of time to wait for a pulse to complete if the device the interface is connected to requires a pin to be 'pulsed' from low to high to low for it to accept data or a command
PULSE_TIME = 1e-6 * 50


class i2c(object):
    """
    Wrap an `I²C <https://en.wikipedia.org/wiki/I%C2%B2C>`_ (Inter-Integrated
    Circuit) interface to provide :py:func:`data` and :py:func:`command` methods.

    :param bus: A *smbus* implementation, if ``None`` is supplied (default),
        `smbus2 <https://pypi.python.org/pypi/smbus2>`_ is used.
        Typically this is overridden in tests, or if there is a specific
        reason why `pysmbus <https://pypi.python.org/pypi/pysmbus>`_ must be used
        over smbus2.
    :type bus:
    :param port: I²C port number, usually 0 or 1 (default).
    :type port: int
    :param address: I²C address, default: ``0x3C``.
    :type address: int
    :raises luma.core.error.DeviceAddressError: I2C device address is invalid.
    :raises luma.core.error.DeviceNotFoundError: I2C device could not be found.
    :raises luma.core.error.DevicePermissionError: Permission to access I2C device
        denied.

    .. note::
       1. Only one of ``bus`` OR ``port`` arguments should be supplied;
          if both are, then ``bus`` takes precedence.
       2. If ``bus`` is provided, there is an implicit expectation
          that it has already been opened.
    """
    def __init__(self, bus=None, port=1, address=0x3C):
        import smbus2
        self._cmd_mode = 0x00
        self._data_mode = 0x40

        try:
            self._addr = int(str(address), 0)
        except ValueError:
            raise luma.core.error.DeviceAddressError(
                'I2C device address invalid: {}'.format(address))

        try:
            self._managed = bus is None
            self._i2c_msg_write = smbus2.i2c_msg.write if bus is None else None
            self._bus = bus or smbus2.SMBus(port)
        except (IOError, OSError) as e:
            if e.errno == errno.ENOENT:
                # FileNotFoundError
                raise luma.core.error.DeviceNotFoundError(
                    'I2C device not found: {}'.format(e.filename))
            elif e.errno in [errno.EPERM, errno.EACCES]:
                # PermissionError
                raise luma.core.error.DevicePermissionError(
                    'I2C device permission denied: {}'.format(e.filename))
            else:  # pragma: no cover
                raise

    def command(self, *cmd):
        """
        Sends a command or sequence of commands through to the I²C address
        - maximum allowed is 32 bytes in one go.

        :param cmd: A spread of commands.
        :type cmd: int
        :raises luma.core.error.DeviceNotFoundError: I2C device could not be found.
        """
        assert(len(cmd) <= 32)

        try:
            self._bus.write_i2c_block_data(self._addr, self._cmd_mode,
                                           list(cmd))
        except (IOError, OSError) as e:
            if e.errno in [errno.EREMOTEIO, errno.EIO]:
                # I/O error
                raise luma.core.error.DeviceNotFoundError(
                    'I2C device not found on address: 0x{0:02X}'.format(self._addr))
            else:  # pragma: no cover
                raise

    def data(self, data):
        """
        Sends a data byte or sequence of data bytes to the I²C address.
        If the bus is in managed mode backed by smbus2, the i2c_rdwr
        method will be used to avoid having to send in chunks.
        For SMBus devices the maximum allowed in one transaction is
        32 bytes, so if data is larger than this, it is sent in chunks.

        :param data: A data sequence.
        :type data: list, bytearray
        """

        # block size is the maximum data payload that will be tolerated.
        # The managed i2c will transfer blocks of upto 4K (using i2c_rdwr)
        # whereas we must use the default 32 byte block size when unmanaged
        if self._managed:
            block_size = 4096
            write = self._write_large_block
        else:
            block_size = 32
            write = self._write_block

        i = 0
        n = len(data)
        while i < n:
            write(list(data[i:i + block_size]))
            i += block_size

    def _write_block(self, data):
        assert len(data) <= 32
        self._bus.write_i2c_block_data(self._addr, self._data_mode, data)

    def _write_large_block(self, data):
        assert len(data) <= 4096
        self._bus.i2c_rdwr(self._i2c_msg_write(self._addr, [self._data_mode] + data))

    def cleanup(self):
        """
        Clean up I²C resources
        """
        if self._managed:
            self._bus.close()


@lib.rpi_gpio
class bitbang(object):
    """
    Wraps an `SPI <https://en.wikipedia.org/wiki/Serial_Peripheral_Interface_Bus>`_
    (Serial Peripheral Interface) bus to provide :py:func:`data` and
    :py:func:`command` methods. This is a software implementation and is thus
    a lot slower than the default SPI interface. Don't use this class directly
    unless there is a good reason!

    :param gpio: GPIO interface (must be compatible with `RPi.GPIO <https://pypi.python.org/pypi/RPi.GPIO>`_).
        For slaves that don't need reset or D/C functionality, supply a
        :py:class:`noop` implementation instead.
    :param transfer_size: Max bytes to transfer in one go. Some implementations
        only support maximum of 64 or 128 bytes, whereas RPi/py-spidev supports
        4096 (default).
    :type transfer_size: int
    :param reset_hold_time: The number of seconds to hold reset active. Some devices may require
        a duration of 100ms or more to fully reset the display (default:0)
    :type reset_hold_time: float
    :param reset_release_time: The number of seconds to delay afer reset. Some devices may require
        a duration of 150ms or more after reset was triggered before the device can accept the
        initialization sequence (default:0)
    :type reset_release_time: float
    :param SCLK: The GPIO pin to connect the SPI clock to.
    :type SCLK: int
    :param SDA: The GPIO pin to connect the SPI data (MOSI) line to.
    :type SDA: int
    :param CE: The GPIO pin to connect the SPI chip enable (CE) line to.
    :type CE: int
    :param DC: The GPIO pin to connect data/command select (DC) to.
    :type DC: int
    :param RST: The GPIO pin to connect reset (RES / RST) to.
    :type RST: int
    """
    def __init__(self, gpio=None, transfer_size=4096, reset_hold_time=0, reset_release_time=0, **kwargs):

        self._transfer_size = transfer_size
        self._managed = gpio is None
        self._gpio = gpio or self.__rpi_gpio__()

        self._SCLK = self._configure(kwargs.get("SCLK"))
        self._SDA = self._configure(kwargs.get("SDA"))
        self._CE = self._configure(kwargs.get("CE"))
        self._DC = self._configure(kwargs.get("DC"))
        self._RST = self._configure(kwargs.get("RST"))
        self._cmd_mode = self._gpio.LOW  # Command mode = Hold low
        self._data_mode = self._gpio.HIGH  # Data mode = Pull high

        if self._RST is not None:
            self._gpio.output(self._RST, self._gpio.LOW)  # Reset device
            sleep(reset_hold_time)
            self._gpio.output(self._RST, self._gpio.HIGH)  # Keep RESET pulled high
            sleep(reset_release_time)

    def _configure(self, pin):
        if pin is not None:
            self._gpio.setup(pin, self._gpio.OUT)
            return pin

    def command(self, *cmd):
        """
        Sends a command or sequence of commands through to the SPI device.

        :param cmd: A spread of commands.
        :type cmd: int
        """
        if self._DC:
            self._gpio.output(self._DC, self._cmd_mode)

        self._write_bytes(list(cmd))

    def data(self, data):
        """
        Sends a data byte or sequence of data bytes through to the SPI device.
        If the data is more than :py:attr:`transfer_size` bytes, it is sent in chunks.

        :param data: A data sequence.
        :type data: list, bytearray
        """
        if self._DC:
            self._gpio.output(self._DC, self._data_mode)

        i = 0
        n = len(data)
        tx_sz = self._transfer_size
        while i < n:
            self._write_bytes(data[i:i + tx_sz])
            i += tx_sz

    def _write_bytes(self, data):
        gpio = self._gpio
        if self._CE:
            gpio.output(self._CE, gpio.LOW)  # Active low

        for byte in data:
            for _ in range(8):
                gpio.output(self._SDA, byte & 0x80)
                gpio.output(self._SCLK, gpio.HIGH)
                byte <<= 1
                gpio.output(self._SCLK, gpio.LOW)

        if self._CE:
            gpio.output(self._CE, gpio.HIGH)

    def cleanup(self):
        """
        Clean up GPIO resources if managed.
        """
        if self._managed:
            self._gpio.cleanup()


@lib.spidev
class spi(bitbang):
    """
    Wraps an `SPI <https://en.wikipedia.org/wiki/Serial_Peripheral_Interface_Bus>`_
    (Serial Peripheral Interface) bus to provide :py:func:`data` and
    :py:func:`command` methods.

    :param spi: SPI implementation (must be compatible with `spidev <https://pypi.python.org/pypi/spidev/>`_)
    :param gpio: GPIO interface (must be compatible with `RPi.GPIO <https://pypi.python.org/pypi/RPi.GPIO>`_).
        For slaves that don't need reset or D/C functionality, supply a
        :py:class:`noop` implementation instead.
    :param port: SPI port, usually 0 (default) or 1.
    :type port: int
    :param device: SPI device, usually 0 (default) or 1.
    :type device: int
    :param bus_speed_hz: SPI bus speed, defaults to 8MHz.
    :type bus_speed_hz: int
    :param cs_high: Whether SPI chip select is high, defaults to ``False``.
    :type cs_high: bool
    :param transfer_size: Maximum amount of bytes to transfer in one go. Some implementations
        only support a maximum of 64 or 128 bytes, whereas RPi/py-spidev supports
        4096 (default).
    :type transfer_size: int
    :param gpio_DC: The GPIO pin to connect data/command select (DC) to (defaults to 24).
    :type gpio_DC: int
    :param gpio_RST: The GPIO pin to connect reset (RES / RST) to (defaults to 25).
    :type gpio_RST: int
    :param spi_mode: SPI mode as two bit pattern of clock polarity and phase [CPOL|CPHA], 0-3 (default:None)
    :type spi_mode: int
    :param reset_hold_time: The number of seconds to hold reset active. Some devices may require
        a duration of 100ms or more to fully reset the display (default:0)
    :type reset_hold_time: float
    :param reset_release_time: The number of seconds to delay afer reset. Some devices may require
        a duration of 150ms or more after reset was triggered before the device can accept the
        initialization sequence (default:0)
    :type reset_release_time: float
    :raises luma.core.error.DeviceNotFoundError: SPI device could not be found.
    :raises luma.core.error.UnsupportedPlatform: GPIO access not available.
    """
    def __init__(self, spi=None, gpio=None, port=0, device=0,
                 bus_speed_hz=8000000, cs_high=False, transfer_size=4096,
                 gpio_DC=24, gpio_RST=25, spi_mode=None,
                 reset_hold_time=0, reset_release_time=0):
        assert(bus_speed_hz in [mhz * 1000000 for mhz in [0.5, 1, 2, 4, 8, 16, 32]])

        bitbang.__init__(self, gpio, transfer_size, reset_hold_time, reset_release_time, DC=gpio_DC, RST=gpio_RST)

        try:
            self._spi = spi or self.__spidev__()
            self._spi.open(port, device)
            if spi_mode:
                self._spi.mode = spi_mode
            self._spi.cshigh = cs_high
        except (IOError, OSError) as e:
            if e.errno == errno.ENOENT:
                raise luma.core.error.DeviceNotFoundError('SPI device not found')
            else:  # pragma: no cover
                raise

        self._spi.max_speed_hz = bus_speed_hz

    def _write_bytes(self, data):
        self._spi.writebytes(data)

    def cleanup(self):
        """
        Clean up SPI & GPIO resources.
        """
        self._spi.close()
        super(spi, self).cleanup()


class gpio_cs_spi(spi):
    """
    Wraps the `spi` class to allow the Chip Select to be used with any GPIO pin.
    The gpio pin to use is defined during instantiation with the keyword argument `gpio_CS`.

    Behaviour is otherwise the same, refer to the documentation on the `spi` class
    for information on other parameters and raised exceptions.

    :param gpio_CS: The GPIO pin to connect chip select (CS / CE) to (defaults to None).
    :type gpio_CS: int
    """
    def __init__(self, *args, **kwargs):
        gpio_CS = kwargs.pop("gpio_CS", None)  # Python 2.7 doesn't allow var args and default values at the same time
        super(gpio_cs_spi, self).__init__(*args, **kwargs)

        if gpio_CS:
            self._gpio_CS = gpio_CS
            self._spi.no_cs = True  # disable spidev's handling of the chip select pin
            self._gpio.setup(self._gpio_CS, self._gpio.OUT, initial=self._gpio.LOW if self._spi.cshigh else self._gpio.HIGH)

    def _write_bytes(self, *args, **kwargs):
        if self._gpio_CS:
            self._gpio.output(self._gpio_CS, self._gpio.HIGH if self._spi.cshigh else self._gpio.LOW)

        super(gpio_cs_spi, self)._write_bytes(*args, **kwargs)

        if self._gpio_CS:
            self._gpio.output(self._gpio_CS, self._gpio.LOW if self._spi.cshigh else self._gpio.HIGH)


class noop(object):
    """
    Does nothing, used for pseudo-devices / emulators / anything really.
    """
    def __getattr__(self, attr):
        return self.__noop

    def __setattr__(self, attr, val):  # pragma: no cover
        pass

    def __noop(self, *args, **kwargs):
        pass


def _ftdi_pin(pin):
    return 1 << pin


class __FTDI_WRAPPER_SPI:
    """
    Adapter for FTDI to spidev. Not for direct public consumption
    """
    def __init__(self, controller, spi_port):
        self._controller = controller
        self._spi_port = spi_port

    def open(self, port, device):
        pass

    def writebytes(self, data):
        self._spi_port.write(data)

    def close(self):
        self._controller.terminate()


class __FTDI_WRAPPER_GPIO:
    """
    Adapter for FTDI to RPI.GPIO. Not for direct public consumption
    """
    LOW = 0
    HIGH = OUT = 1

    def __init__(self, gpio):
        self._gpio = gpio
        self._data = 0

    def setup(self, pin, direction):
        pass

    def output(self, pin, value):
        mask = _ftdi_pin(pin)
        self._data &= ~mask
        if value:
            self._data |= mask

        self._gpio.write(self._data)

    def cleanup(self):
        pass


class __FTDI_WRAPPER_I2C:
    """
    Adapter for FTDI to I2C smbus. Not for direct public consumption
    """

    def __init__(self, controller, i2c_port):
        self._controller = controller
        self._i2c_port = i2c_port

    def write_i2c_block_data(self, address, register, data):
        self._i2c_port.write_to(register, data)

    def i2c_rdwr(self, message):
        address, data = message
        register = data[0]
        self.write_i2c_block_data(address, register, data[1:])

    def close(self):
        self._controller.terminate()


def ftdi_spi(device='ftdi://::/1', bus_speed_hz=12000000, gpio_CS=3, gpio_DC=5, gpio_RST=6,
        reset_hold_time=0, reset_release_time=0):
    """
    Bridges an `SPI <https://en.wikipedia.org/wiki/Serial_Peripheral_Interface_Bus>`_
    (Serial Peripheral Interface) bus over an FTDI USB device to provide :py:func:`data` and
    :py:func:`command` methods.

    :param device: A URI describing the location of the FTDI device. If ``None`` is
        supplied (default), ``ftdi://::/1`` is used. See `pyftdi <https://pypi.python.org/pypi/pyftdi>`_
        for further details of the naming scheme used.
    :type device: string
    :param bus_speed_hz: SPI bus speed, defaults to 12MHz.
    :type bus_speed_hz: int
    :param gpio_CS: The ADx pin to connect chip select (CS) to (defaults to 3).
    :type gpio_CS: int
    :param gpio_DC: The ADx pin to connect data/command select (DC) to (defaults to 5).
    :type gpio_DC: int
    :param gpio_RST: The ADx pin to connect reset (RES / RST) to (defaults to 6).
    :type gpio_RST: int
        :param reset_hold_time: The number of seconds to hold reset active. Some devices may require
        a duration of 100ms or more to fully reset the display (default:0)
    :type reset_hold_time: float
    :param reset_release_time: The number of seconds to delay afer reset. Some devices may require
        a duration of 150ms or more after reset was triggered before the device can accept the
        initialization sequence (default:0)
    :type reset_release_time: float

    .. versionadded:: 1.9.0
    """
    from pyftdi.spi import SpiController

    controller = SpiController(cs_count=1)
    controller.configure(device)

    slave = controller.get_port(cs=gpio_CS - 3, freq=bus_speed_hz, mode=0)
    gpio = controller.get_gpio()

    # RESET and DC configured as outputs
    pins = _ftdi_pin(gpio_RST) | _ftdi_pin(gpio_DC)
    gpio.set_direction(pins, pins & 0xFF)

    serial = spi(
        __FTDI_WRAPPER_SPI(controller, slave),
        __FTDI_WRAPPER_GPIO(gpio),
        gpio_DC=gpio_DC,
        gpio_RST=gpio_RST,
        reset_hold_time=reset_hold_time,
        reset_release_time=reset_release_time)
    serial._managed = True
    return serial


def ftdi_i2c(device='ftdi://::/1', address=0x3C):
    """
    Bridges an `I²C <https://en.wikipedia.org/wiki/I%C2%B2C>`_ (Inter-Integrated
    Circuit) interface over an FTDI USB device to provide :py:func:`data` and
    :py:func:`command` methods.

    :param device: A URI describing the location of the FTDI device. If ``None`` is
        supplied (default), ``ftdi://::/1`` is used. See `pyftdi <https://pypi.python.org/pypi/pyftdi>`_
        for further details of the naming scheme used.
    :type device: string
    :param address: I²C address, default: ``0x3C``.
    :type address: int
    :raises luma.core.error.DeviceAddressError: I2C device address is invalid.

    .. versionadded:: 1.9.0
    """
    from pyftdi.i2c import I2cController

    try:
        addr = int(str(address), 0)
    except ValueError:
        raise luma.core.error.DeviceAddressError(
            'I2C device address invalid: {}'.format(address))

    controller = I2cController()
    controller.configure(device)

    port = controller.get_port(addr)

    serial = i2c(bus=__FTDI_WRAPPER_I2C(controller, port))
    serial._managed = True
    serial._i2c_msg_write = lambda address, data: (address, data)
    return serial


class pcf8574(i2c):
    """
    I²C interface to provide :py:func:`data` and :py:func:`command` methods
    for a device using a pcf8574 backpack.

    :param bus: A *smbus* implementation, if ``None`` is supplied (default),
        `smbus2 <https://pypi.python.org/pypi/smbus2>`_ is used.
        Typically this is overridden in tests, or if there is a specific
        reason why `pysmbus <https://pypi.python.org/pypi/pysmbus>`_ must be used
        over smbus2.
    :type bus:
    :param port: I²C port number, usually 0 or 1 (default).
    :type port: int
    :param address: I²C address, default: ``0x3C``.
    :type address: int
    :param pulse_time: length of time in seconds that the enable line should be
        held high during a data or command transfer (default: 50μs)
    :type pulse_time: float
    :param backlight_enabled: Determines whether to activate the display's backlight
    :type backlight_enabled: bool
    :param RS: where register/select is connected to the backpack (default: 0)
    :type RS: int
    :param E: where enable pin is connected to the backpack (default: 2)
    :type E: int
    :param PINS: The PCF8574 pins that form the data bus in LSD to MSD order
    :type PINS: list[int]
    :param BACKLIGHT: Pin number of the pcf8574 (counting from zero) that the
        backlight is controlled from (default: 3)
    :type BACKLIGHT: int
    :param COMMAND: determines whether RS high sets device to expect a command
        byte or a data byte.  Must be either ``high`` (default) or ``low``
    :type COMMAND: str

    :raises luma.core.error.DeviceAddressError: I2C device address is invalid.
    :raises luma.core.error.DeviceNotFoundError: I2C device could not be found.
    :raises luma.core.error.DevicePermissionError: Permission to access I2C device
        denied.

    .. note::
       1. Only one of ``bus`` OR ``port`` arguments should be supplied;
          if both are, then ``bus`` takes precedence.
       2. If ``bus`` is provided, there is an implicit expectation
          that it has already been opened.
       3. Default wiring:

       * RS - Register Select
       * E - Enable
       * RW - Read/Write (note: unused by this driver)
       * D4-D7 - The upper data pins

       ========= === === === === === === === =========
       Device     RS  RW   E  D4  D5  D6  D7 BACKLIGHT
       Display     4   5   6  11  12  13  14
       Backpack   P0  P1  P2  P4  P5  P6  P7        P3
       ========= === === === === === === === =========

       If your PCF8574 is wired up differently to this you will need to provide
       the correct values for the RS, E, COMMAND, BACKLIGHT parameters.
       RS, E and BACKLIGHT are set to the pin numbers of the backpack pins
       they are connect to from P0-P7.

       COMMAND is set to 'high' if the Register Select (RS) pin needs to be high
       to inform the device that a command byte is being sent or 'low' if RS low
       is used for commands.

       PINS is a list of the pin positions that match where the devices data
       pins have been connected on the backpack (P0-P7).  For many devices this
       will be d4->P4, d5->P5, d6->P6, and d7->P7 ([4, 5, 6, 7]) which is the
       default.

       Example:

       If your data lines D4-D7 are connected to the PCF8574s pins P0-P3 with
       the RS pin connected to P4, the enable pin to P5, the backlight pin
       connected to P7, and the RS value to indicate command is low, your
       initialization would look something like:

       ``pcf8574(port=1, address=0x27, PINS=[0, 1, 2, 3], RS=4, E=5,
       COMMAND='low', BACKLIGHT=7)``

       Explanation:
       PINS are set to ``[0, 1, 2, 3]`` which assigns P0 to D4, P1 to D5, P2 to D6,
       and P3 to D7.  RS is set to 4 to associate with P4. Similarly E is set
       to 5 to associate E with P5.  BACKLIGHT set to 7 connects it to pin P7
       of the backpack.  COMMAND is set to ``low`` so that RS will be set to low
       when a command is sent and high when data is sent.

    .. versionadded:: 1.15.0
    """

    _BACKLIGHT = 3
    _ENABLE = 2
    _RS = 0
    _OFFSET = 4
    _CMD = 'low'

    def __init__(self, pulse_time=PULSE_TIME, backlight_enabled=True, *args, **kwargs):
        super(pcf8574, self).__init__(*args, **kwargs)

        self._pulse_time = pulse_time
        self._bitmode = 4  # PCF8574 can only be used to transfer 4 bits at a time

        self._PINS = kwargs.get('PINS', list((4, 5, 6, 7)))
        self._datalines = len(self._PINS)
        assert self._datalines == 4, 'You\'ve provided {0} data pins but the PCF8574 only supports four'.format(len(self._PINS))

        self._rs = self._mask(kwargs.get("RS", self._RS))
        self._cmd = 0xFF if kwargs.get("COMMAND", self._CMD).lower() == 'high' else 0x00
        self._data = 0x00 if self._cmd else 0xFF
        self._cmd_mode = self._rs & self._cmd
        self._data_mode = self._rs & self._data
        self._enable = self._mask(kwargs.get("ENABLE", self._ENABLE))
        self._backlight_enabled = self._mask(kwargs.get("BACKLIGHT", self._BACKLIGHT)) if backlight_enabled else 0x00

    def command(self, *cmd):
        """
        Sends a command or sequence of commands through to the I²C address
        - maximum allowed is 32 bytes in one go.

        :param cmd: A spread of commands in high_bits, low_bits order.
        :type cmd: int
        :raises luma.core.error.DeviceNotFoundError: I2C device could not be found.

        IMPORTANT: the PCF8574 only supports four bit transfers.  It is the
        devices responsibility to break each byte sent into a high bit
        and a low bit transfer.

        Example:
            To set an HD44780s cursor to the beginning of the first line requires
            sending 0b10000000 (0x80).  This is 0b1000 (0x08) at the high side of
            the byte and 0b0000 (0x00) on the low side of the byte.

            For example, to send this using the pcf8574 interface::

              d = pcf8574(bus=1, address=0x27)
              d.command([0x08, 0x00])
        """
        self._write(list(cmd), self._cmd_mode)

    def data(self, data):
        """
        Sends a data byte or sequence of data bytes to the I²C address.

        :param data: A data sequence.
        :type data: list, bytearray

        IMPORTANT: the PCF8574 only supports four bit transfers.  It is the
        devices responsibility to break each byte sent into a high bit
        and a low bit transfer.

        Example:
            To send an ascii 'A' (0x41) to the display you need to send binary
            01000001.  This is 0100 (0x40) at the high side of the byte
            and 0001 (0x01) on the low side of the byte.

            For example, to send this using the pcf8574 interface::

              d = pcf8574(bus=1, address=0x27)
              d.command([0x04, 0x01])
        """
        self._write(data, self._data_mode)

    def _mask(self, pin):
        """
        Return a mask that contains a 1 in the pin position
        """
        return 1 << pin

    def _compute_pins(self, value):
        """
        Set bits in value according to the assigned pin positions on the PCF8574
        """
        retv = 0
        for i in range(self._datalines):
            retv |= ((value >> i) & 0x01) << self._PINS[i]
        return retv

    def _write(self, data, mode):
        try:
            for value in data:
                self._bus.write_byte(self._addr, self._backlight_enabled | mode | self._compute_pins(value))
                self._bus.write_byte(self._addr, self._backlight_enabled | mode | self._compute_pins(value) | self._enable)
                sleep(self._pulse_time)
                self._bus.write_byte(self._addr, self._backlight_enabled | mode | self._compute_pins(value))
        except (IOError, OSError) as e:
            if e.errno in [errno.EREMOTEIO, errno.EIO]:
                # I/O error
                raise luma.core.error.DeviceNotFoundError(
                    'I2C device not found on address: 0x{0:02X}'.format(self._addr))
            else:  # pragma: no cover
                raise



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Encapsulates sending commands and data over a parallel-bus interface.
"""

from time import sleep
from luma.core import lib


__all__ = ["bitbang_6800"]

'''
Default amount of time to wait for a pulse to complete if the device the
interface is connected to requires a pin to be 'pulsed' from low to high to
low for it to accept data or a command
'''
PULSE_TIME = 1e-6 * 50


@lib.rpi_gpio
class bitbang_6800(object):
    """
    Implements a 6800 style parallel-bus interface that provides :py:func:`data`
    and :py:func:`command` methods. The default pin assignments provided are
    from `Adafruit <https://learn.adafruit.com/drive-a-16x2-lcd-directly-with-a-raspberry-pi/wiring>`_.

    :param gpio: GPIO interface (must be compatible with
        `RPi.GPIO <https://pypi.python.org/pypi/RPi.GPIO>`_)
    :param pulse_time: length of time in seconds that the enable line should be
        held high during a data or command transfer
    :type pulse_time: float
    :param RS: The GPIO pin register select (RS) line (low for command, high
        for data)
    :type RS: int
    :param E: The GPIO pin to connect the enable (E) line to.
    :type E: int
    :param PINS: The GPIO pins that form the data bus (a list of 4 or 8 pins
        depending upon implementation ordered from LSD to MSD)
    :type PINS: list[int]

    .. versionadded:: 1.16.2
    """

    def __init__(self, gpio=None, pulse_time=PULSE_TIME, **kwargs):
        self._managed = gpio is None
        self._gpio = gpio or self.__rpi_gpio__()
        self._gpio.setwarnings(False)
        self._pulse_time = pulse_time

        self._RS = self._configure(kwargs.get("RS", 22))
        self._E = self._configure(kwargs.get("E", 17))
        self._PINS = self._configure(kwargs.get('PINS', list((25, 24, 23, 18))))

        self._datalines = len(self._PINS)
        assert self._datalines in (4, 8), 'You\'ve provided {0} pins but a bus must contain either four or eight pins'.format(len(self._PINS))
        self._bitmode = self._datalines  # Used by device to autoset its own bitmode

        self._cmd_mode = self._gpio.LOW  # Command mode = Hold low
        self._data_mode = self._gpio.HIGH  # Data mode = Pull high

    def _configure(self, pin):
        pins = pin if type(pin) == list else [pin] if pin else []
        for p in pins:
            self._gpio.setup(p, self._gpio.OUT)
        return pin

    def command(self, *cmd):
        """
        Sends a command or sequence of commands through the bus
        If the bus is in four bit mode, only the lowest 4 bits of the data
        value will be sent.

        This means that the device needs to send high and low bits separately
        if the device is operating using a 4 bit bus (e.g. to send a 0x32 in
        4 bit mode the device would use ``command(0x03, 0x02)``).

        :param cmd: A spread of commands.
        :type cmd: int
        """
        self._write(list(cmd), self._cmd_mode)

    def data(self, data):
        """
        Sends a data byte or sequence of data bytes through to the bus.
        If the bus is in four bit mode, only the lowest 4 bits of the data
        value will be sent.

        This means that the device needs to send high and low bits separately
        if the device is operating using a 4 bit bus (e.g. to send a 0x32 in
        4 bit mode the device would use ``data([0x03, 0x02])``).

        :param data: A data sequence.
        :type data: list, bytearray
        """
        self._write(data, self._data_mode)

    def _write(self, data, mode):
        gpio = self._gpio
        gpio.output(self._RS, mode)
        gpio.output(self._E, gpio.LOW)
        for value in data:
            for i in range(self._datalines):
                gpio.output(self._PINS[i], (value >> i) & 0x01)
            gpio.output(self._E, gpio.HIGH)
            sleep(self._pulse_time)
            gpio.output(self._E, gpio.LOW)

    def cleanup(self):
        """
        Clean up GPIO resources if managed.
        """
        if self._managed:
            self._gpio.cleanup()



# -*- coding: utf-8 -*-
# Copyright (c) 2017-20 Richard Hull and contributors
# See LICENSE.rst for details.

import atexit
from time import sleep

from luma.core import mixin
from luma.core.util import bytes_to_nibbles
import luma.core.const
from luma.core.interface.serial import i2c, noop


class device(mixin.capabilities):
    """
    Base class for display driver classes

    .. note::
        Direct use of the :func:`command` and :func:`data` methods are
        discouraged: Screen updates should be effected through the
        :func:`display` method, or preferably with the
        :class:`luma.core.render.canvas` context manager.
    """
    def __init__(self, const=None, serial_interface=None):
        self._const = const or luma.core.const.common
        self._serial_interface = serial_interface or i2c()

        def shutdown_hook():  # pragma: no cover
            try:
                self.cleanup()
            except:
                pass

        atexit.register(shutdown_hook)

    def command(self, *cmd):
        """
        Sends a command or sequence of commands through to the delegated
        serial interface.
        """
        self._serial_interface.command(*cmd)

    def data(self, data):
        """
        Sends a data byte or sequence of data bytes through to the delegated
        serial interface.
        """
        self._serial_interface.data(data)

    def show(self):
        """
        Sets the display mode ON, waking the device out of a prior
        low-power sleep mode.
        """
        self.command(self._const.DISPLAYON)

    def hide(self):
        """
        Switches the display mode OFF, putting the device in low-power
        sleep mode.
        """
        self.command(self._const.DISPLAYOFF)

    def contrast(self, level):
        """
        Switches the display contrast to the desired level, in the range
        0-255. Note that setting the level to a low (or zero) value will
        not necessarily dim the display to nearly off. In other words,
        this method is **NOT** suitable for fade-in/out animation.

        :param level: Desired contrast level in the range of 0-255.
        :type level: int
        """
        assert(0 <= level <= 255)
        self.command(self._const.SETCONTRAST, level)

    def cleanup(self):
        """
        Attempt to switch the device off or put into low power mode (this
        helps prolong the life of the device), clear the screen and close
        resources associated with the underlying serial interface.

        If :py:attr:`persist` is ``True``, the device will not be switched off.

        This is a managed function, which is called when the python processs
        is being shutdown, so shouldn't usually need be called directly in
        application code.
        """
        if not self.persist:
            self.hide()
            self.clear()
        self._serial_interface.cleanup()


class parallel_device(device):
    """
    Wrapper class to manage communications with devices that can operate in
    four or eight bit modes.

    .. note::
        parallel_devices require specific timings which are managed by using
        ``time.sleep`` to cause the process to block for small amounts of time.
        If your application is especially time sensitive, consider running the
        drivers in a separate thread.

    .. versionadded:: 1.16.0
    """

    def __init__(self, const=None, serial_interface=None, exec_time=None, **kwargs):
        super(parallel_device, self).__init__(const, serial_interface)

        self._exec_time = exec_time if exec_time is not None else \
            serial_interface._pulse_time if hasattr(serial_interface, '_pulse_time') \
            else 0
        self._bitmode = serial_interface._bitmode if hasattr(serial_interface, '_bitmode') else 4
        assert self._bitmode in (4, 8), 'Bit mode {0} is invalid.  It can only be 4 or 8'.format(self._bitmode)

    def command(self, *cmd, exec_time=None, only_low_bits=False):
        """
        Sends a command or sequence of commands through to the serial interface.
        If operating in four bit mode, expands each command from one byte
        values (8 bits) to two nibble values (4 bits each)

        :param cmd: A spread of commands.
        :type cmd: int
        :param exec_time: Amount of time to wait for the command to finish
            execution.  If not provided, the device default will be used instead
        :type exec_time: float
        :param only_low_bits: If ``True``, only the lowest four bits of the command
            will be sent.  This is necessary on some devices during initialization
        :type only_low_bits: bool
        """
        cmd = cmd if (self._bitmode == 8 or only_low_bits) else \
            bytes_to_nibbles(cmd)
        super(parallel_device, self).command(*cmd)
        sleep(exec_time or self._exec_time)

    def data(self, data):
        """
        Sends a sequence of bytes through to the serial interface.
        If operating in four bit mode, expands each byte from a single
        value (8 bits) to two nibble values (4 bits each)

        :param data: a sequence of bytes to send to the display
        :type data: list
        """
        data = data if self._bitmode == 8 else \
            bytes_to_nibbles(data)
        super(parallel_device, self).data(data)


class dummy(device):
    """
    Pseudo-device that acts like a physical display, except that it does nothing
    other than retain a copy of the displayed image. It is mostly useful for
    testing. Supports 24-bit color depth.
    """
    def __init__(self, width=128, height=64, rotate=0, mode="RGB", **kwargs):
        super(dummy, self).__init__(serial_interface=noop())
        self.capabilities(width, height, rotate, mode)
        self.image = None

    def display(self, image):
        """
        Takes a :py:mod:`PIL.Image` and makes a copy of it for later
        use/inspection.

        :param image: Image to display.
        :type image: PIL.Image.Image
        """
        assert(image.size == self.size)

        self.image = self.preprocess(image).copy()
